package function.builtin.special;

import static function.builtin.EVAL.evalRecurArgumentList;

import error.LispException;
import function.ArgumentValidator;
import function.FunctionNames;
import function.LispFunction;
import function.LispSpecialFunction;
import sexpression.Cons;
import sexpression.SExpression;
import table.ExecutionContext;

@FunctionNames({ "RECUR" })
public class RECUR extends LispSpecialFunction {

    private ArgumentValidator argumentValidator;
    private ExecutionContext executionContext;

    public RECUR(String name) {
        this.argumentValidator = new ArgumentValidator(name);
        this.executionContext = ExecutionContext.getInstance();
    }

    @Override
    public SExpression call(Cons argumentList) {
        if (executionContext.isRecur())
            throw new NestedRecurException();

        if (!executionContext.isInFunctionCall())
            throw new RecurOutsideOfFunctionException();

        argumentValidator.validate(argumentList);

        executionContext.setRecur();
        LispFunction currentFunction = executionContext.getCurrentFunction();
        Cons recurArguments = argumentList;

        if (currentFunction.isArgumentListEvaluated())
            recurArguments = evalRecurArgumentList(argumentList);

        return recurArguments;
    }

    public static class RecurOutsideOfFunctionException extends LispException {

        private static final long serialVersionUID = 1L;

        @Override
        public String getMessage() {
            return "recur called outide of function";
        }
    }

    public static class NestedRecurException extends LispException {

        private static final long serialVersionUID = 1L;

        @Override
        public String getMessage() {
            return "nested call to recur";
        }
    }

}
